#include "camera_parking.h"

namespace
{
    std::vector<mdc::camera::CameraAccessConfig> camera_configs = {
        {
            .moduleName = "HUAWEI_BU18TM41_OX03C10",
#ifdef CONF_SOC_A
            .cameraId = 8U,
            .fov = 200U,
#else
            .cameraId = 4U,
            .fov = 200U,
#endif
            .fps = 20,
            .width = 1920U,
            .height = 1080U,
            .streamDepth = 5U,
            .clusterId = 0U,

        },
        {
            .moduleName = "HUAWEI_BU18TM41_OX03C10",
#ifdef CONF_SOC_A
            .cameraId = 9U,
            .fov = 200U,
#else
            .cameraId = 5U,
            .fov = 200U,
#endif
            .fps = 20,
            .width = 1920U,
            .height = 1080U,
            .streamDepth = 3U,
            .clusterId = 1U,
        },
        {
            .moduleName = "HUAWEI_BU18TM41_OX03C10",
#ifdef CONF_SOC_A
            .cameraId = 10U,
            .fov = 200U,
#else
            .cameraId = 6U,
            .fov = 200U,
#endif
            .fps = 20,
            .width = 1920U,
            .height = 1080U,
            .streamDepth = 5U,
            .clusterId = 2U,
        },
        {
            .moduleName = "HUAWEI_BU18TM41_OX03C10",
#ifdef CONF_SOC_A
            .cameraId = 11U,
            .fov = 200U,
#else
            .cameraId = 7U,
            .fov = 200U,
#endif
            .fps = 20,
            .width = 1920U,
            .height = 1080U,
            .streamDepth = 3U,
            .clusterId = 3U,
        }};
}

//CameraParking::CameraParking() : _encoderWrapper("camera_prk/camera_prk/CamPrk2VisualizationPPort", this->logger), _perceptionWrapper("camera_prk/camera_prk/CamPrk2CamPrkPerceptionPPort", this->logger)
//{
//}
CameraParking::CameraParking()
{
}
CameraParking::~CameraParking()
{
    delete this->camera;
}

void CameraParking::init(unsigned int inputImageWidth, unsigned int inputImageHeight, unsigned int outputImageWidth, unsigned int outputImageHeight, unsigned int cropTopOffset, unsigned int cropLeftOffset, unsigned int encoderGopNumber)
{
    this->inputImageWidth = inputImageWidth;
    this->inputImageHeight = inputImageHeight;
    this->outputImageWidth = outputImageWidth;
    this->outputImageHeight = outputImageHeight;
    this->cropTopOffset = cropTopOffset;
    this->cropLeftOffset = cropLeftOffset;
    this->encoderGopNumber = encoderGopNumber;
    for (auto &temp : this->encodeImagePublishedCounter)
    {
        temp = 0;
    }

    this->mbufManager.init(true, 20U, (this->outputImageWidth * this->outputImageHeight * 3));
    //this->_encoderWrapper.Init();
    //this->_perceptionWrapper.Init();

    // this->sendBanner();
    this->framePool = std::make_shared<CameraPool>();
    this->camera = new MDCCamera(this->framePool);
    this->camera->init(camera_configs);
    for (uint8_t i = 0; i < 4; i++)
    {
        this->pipelineThreadPool[i] = std::thread(&CameraParking::threadsFunction, this, i);
        if (i == 0)
        {
            std::this_thread::sleep_for(std::chrono::milliseconds(1000));
        }
    }
    for (auto &t : this->pipelineThreadPool)
    {
        t.join();
    }
}

void CameraParking::sendBanner()
{
    size_t bannerDataLength;
    void *bannerData = load_file("/opt/usr/home/mdc/wks/camera_integration/data/image_for_overlay.h265", bannerDataLength);
    int i = 0;
    timespec ts;
    clock_gettime(CLOCK_VIRTUAL, &ts);
    mdc::visual::Image image;
    image.frameData.assign((uint8_t *)bannerData, (uint8_t *)bannerData + bannerDataLength);
    image.header.frameId = "mdc_camera_instance_71";
    image.header.stamp.sec = ts.tv_sec;
    image.header.stamp.nsec = ts.tv_nsec;
    bool gopFlag;
    while (i++ < 50)
    {
        for (uint8_t j = 0; j < 4; j++)
        {
            gopFlag = this->encodeImagePublishedCounter[j] == this->encoderGopNumber;
            //this->_encoderWrapper.s.publishEncodedImage(image, j, gopFlag);
            //this->_encoderWrapper.Publish();
            this->encodeImagePublishedCounter[j]++;
            std::this_thread::sleep_for(std::chrono::milliseconds(10));
        }
    }
    free(static_cast<char *>(bannerData));
}

void *CameraParking::load_file(std::string filepath, size_t &length)
{
    std::ifstream file(filepath.c_str(), std::ios::binary);

    if (!file.is_open())
    {
        this->logger.LogError() << "cannot open file";
        return nullptr;
    }
    file.seekg(0, std::ios::end);
    length = static_cast<size_t>(file.tellg());
    file.seekg(0, std::ios::beg);
    this->logger.LogInfo() << "length of the file is " << length;
    void *buffer = new char[length];
    file.read(static_cast<char *>(buffer), length);
    file.close();
    return buffer;
}

void CameraParking::threadsFunction(uint8_t id)
{
    CamPreprocess preprocess(id);
    preprocess.setLoggerName("PRE" + std::to_string(id));
    preprocess.init(1, this->inputImageWidth, this->inputImageHeight, this->outputImageWidth, this->outputImageHeight, this->cropTopOffset, this->cropLeftOffset, true, this->encoderGopNumber);
    preprocess.allocateMemory(0);
    std::string name = "test" + std::to_string(id);
    while (true)
    {
        float duration = 0;
        {
            Timer t(name, &duration);
            std::unique_lock<std::mutex> lock(this->framePool->mtx[id]);
            this->framePool->cv[id].wait(lock, [&]
                                     { return this->framePool->dataReady[id]; });
            this->runPipelines(preprocess, id);
            this->framePool->dataReady[id] = false;
        }
        this->logger.LogDebug() << "Runtime " << name.c_str() << ": " << duration << " milliseconds.";
    }
}

void CameraParking::runPipelines(CamPreprocess &processor, uint8_t id)
{
    size_t length;
    uint8_t failed = this->camera->getCameraYUVBuffer(&processor.inputPicInfo[0].picture_address, length, this->framePool->pool[id].imageVector[this->framePool->pool[id].imageVector.size() - 1]);
    if (!failed)
    {
        // encoder pipeline
        std::shared_ptr<hi_venc_stream> output_enc;
        int32_t failed = processor.runEncoderPipeline(0, &output_enc);
        if (!failed)
        {
            std::lock_guard<std::mutex> lock(this->m_writer_mutex1);
            timespec ts;
            clock_gettime(CLOCK_VIRTUAL, &ts);
            mdc::visual::Image image;
            image.frameData.assign((uint8_t *)output_enc->pack[0].addr + output_enc->pack[0].offset, (uint8_t *)output_enc->pack[0].addr + (size_t)output_enc->pack[0].len);
            image.header.frameId = "mdc_camera_instance_71";
            image.header.stamp.sec = ts.tv_sec;
            image.header.stamp.nsec = ts.tv_nsec;
            bool gopFlag = (this->encodeImagePublishedCounter[id] % this->encoderGopNumber == 0);
            if (this->encodeImagePublishedCounter[id] == this->encoderGopNumber)
            {
                this->encodeImagePublishedCounter[id] = 0;
            }
            //this->_encoderWrapper.s.publishEncodedImage(image, id, gopFlag);
            //this->_encoderWrapper.Publish();
            this->encodeImagePublishedCounter[id]++;
            this->logger.LogDebug() << "encoder pipeline completed successfully with id: " << id;
        }
        else
        {
            this->logger.LogError() << "encoder pipeline failed";
        }

        // perception pipeline
        failed = processor.runPerceptionPipeline(0);
        if (!failed)
        {
            std::lock_guard<std::mutex> lock(this->m_writer_mutex2);
            uint64_t mbufDataSize;
            Mbuf *pMbuf = this->mbufManager.fillMbuf(processor.outputPicInfo[0].picture_address, &mbufDataSize, processor.outputPicInfo[0].picture_buffer_size);
            //this->_perceptionWrapper.s.publishPerceptionImage(reinterpret_cast<CameraDrivingBuffer *>(pMbuf), id, mbufDataSize);
            //this->_perceptionWrapper.Publish();
            this->logger.LogDebug() << "perception pipeline completed successfully with id: " << id;
            // uint64_t size;
            // uint8_t *pData = this->mbufManager.getMbufData(&pMbuf, &size);
            // this->logger.LogDebug() << "data in mbuf is: " << pData;
        }
        else
        {
            this->logger.LogError() << "perception pipeline failed";
        }
    }
    else
    {
        this->logger.LogError() << "get YUV buffer failed in camera " << id << " with value 0x" << failed;
    }
    this->framePool->pool[id].imageVector.clear();
}
